"xvalid" <-
  function (geodata, coords = geodata$coords, data = geodata$data, 
            model, trend, reestimate = FALSE, output.reestimate = FALSE, 
            locations.xvalid = "all", messages.screen = TRUE, ...) 
{
  n <- nrow(coords)
  data <- as.vector(data)
  if (length(data) != n) 
    stop("coords and data have imcompatible dimentions")
  xmat <- trend.spatial(trend = trend, coords = coords)
  if(locations.xvalid == "all") locations.xvalid <- 1:n
  else{
    if(is.matrix(locations.xvalid) | is.data.frame(locations.xvalid))
      locations.xvalid <- is.vector(locations.xvalid)
    if(!is.vector(locations.xvalid) | !is.numeric(locations.xvalid))
      stop("argument locations.xvalid must be a numeric vector with numbers indicating the locations to be cross-validated")
    if(any(locations.xvalid) > n | length(locations.xvalid) > n)
      stop("incorrect value to the argument locations.xvalid.\nThis must be a numeric vector with numbers indicating the locations to be cross-validated")
  }
  n.pt.xv <- length(locations.xvalid)
  if(messages.screen){
    cat(paste("xvalid: number of data locations                 =", n))
    cat("\n")
    cat(paste("xvalid: number of locations for cross-validation =", n.pt.xv))
    cat("\n")
    cat("xvalid: performing cross-validation at location:\n")
    }
  cv.f <- function(ndata, ...) {
    if(messages.screen) cat(paste(ndata, ", ", sep=""))
    coords.out <- coords[ndata, , drop = FALSE]
    data.out <- data[ndata]
    xmat.out <- xmat[ndata, , drop = FALSE]
    cv.coords <- coords[-ndata, ]
    cv.data <- as.vector(data)[-ndata]
    cv.xmat <- xmat[-ndata, , drop = FALSE]
    if (reestimate) {
      fix.pars <- (model$parameters.summary[c("tausq", 
                                              "kappa", "psiA", "psiR", "lambda"), 1] == "fixed")
      val.pars <- model$parameters.summary[c("tausq", "kappa", 
                                             "psiA", "psiR", "lambda"), 2]
      names(fix.pars) <- c("tausq", "kappa", "psiA", "psiR", 
                           "lambda")
      names(val.pars) <- c("tausq", "kappa", "psiA", "psiR", 
                           "lambda")
      model <- likfit(coords = cv.coords, data = cv.data, 
                      ini = model$cov.pars, fix.nugget = fix.pars["tausq"], 
                      nugget = val.pars["tausq"], fix.kappa = fix.pars["kappa"], 
                      kappa = val.pars["kappa"], fix.psiR = fix.pars["psiR"], 
                      psiR = val.pars["psiR"], fix.psiA = fix.pars["psiA"], 
                      psiA = val.pars["psiA"], fix.lambda = fix.pars["lambda"], 
                      lambda = val.pars["lambda"], cov.model = model$cov.model, 
                      trend = ~cv.xmat + 0, method = model$method, 
                      mess = F, ...)
    }
    fix.pars <- (model$parameters.summary[c("tausq", "kappa", 
                                            "psiA", "psiR", "lambda"), 1] == "fixed")
    val.pars <- model$parameters.summary[c("tausq", "kappa", 
                                           "psiA", "psiR", "lambda"), 2]
    names(fix.pars) <- c("tausq", "kappa", "psiA", "psiR", 
                         "lambda")
    names(val.pars) <- c("tausq", "kappa", "psiA", "psiR", 
                         "lambda")
    kr <- krige.conv(coords = cv.coords, data = cv.data, 
                     loc = coords.out, krige = krige.control(trend.d = ~cv.xmat + 
                                         0, trend.l = ~xmat.out + 0, cov.model = model$cov.model, 
                                         cov.pars = model$cov.pars, nugget = model$nugget, 
                                         kappa = val.pars["kappa"], lambda = val.pars["lambda"], 
                                         aniso.pars = val.pars[c("psiA", "psiR")]), mess = FALSE)
    res <- c(data.out, kr$pred, kr$krige.var)
    ##, err = (data.out - kr$pred), e.rel = (data.out - kr$pred)/sqrt(kr$krige.var), 
    ##pval = pnorm(data.out, mean = kr$pred, sd = sqrt(kr$krige.var)))
    return(res)
  }
  res <- as.data.frame(t(apply(matrix(locations.xvalid), 1, cv.f)))
  if(messages.screen) cat("\n xvalid: end of cross-validation\n")
  names(res) <- c("data", "pred", "krige.var")
  res$err <- res$data - res$pred
  res$rel.err <- res$err/sqrt(res$krige.var)
  res$pval <- pnorm(res$data, mean = res$pred, sd = sqrt(res$krige.var))  
  return(res)
}
